import asyncio
import json
import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, TypeAlias, Callable, Awaitable

from jinja2 import Environment, Template, FileSystemLoader

from .models.base import BaseArchytasModel

prompt_path = Path(__file__).parent / "prompts"
jinja_env = Environment(loader=FileSystemLoader(prompt_path))

if TYPE_CHECKING:
    from .agent import Agent
    from .chat_history import ChatHistory, MessageRecord, SummaryRecord, RecordType, AutoContextMessage
    from langchain_core.messages import ToolMessage, AIMessage, BaseMessage, HumanMessage, SystemMessage

logger = logging.getLogger("beaker")


MessageSummarizerFunction: TypeAlias = "Callable[[ToolMessage, ChatHistory, Agent, Optional[Model]], Awaitable[None]]"
LoopSummarizerFunction: TypeAlias = "Callable[[list[MessageRecord], ChatHistory, Agent, bool Optional[Model]], Awaitable[None]]"
HistorySummarizerFunction: TypeAlias = "Callable[[ChatHistory, Agent], Awaitable[None]]"

MESSAGE_SUMMARIZATION_THRESHOLD: int = 1000
MESSAGE_SUMMARIZATION_SNIPPET_SIZE: int = 1000


async def get_summarizable_records(chat_history: "ChatHistory") -> "list[RecordType]":
    from .chat_history import RecordType, SummaryRecord, SystemMessage, AutoContextMessage, AIMessage
    all_records = await chat_history.records(auto_update_context=False)
    summarizable_records: list[RecordType] = [
        record for record in all_records
        if record.message is not None
        and (
            isinstance(record, SummaryRecord)
            or not isinstance(record.message, (SystemMessage, AutoContextMessage))
        )
    ]
    return summarizable_records


async def get_records_up_to_threshold(chat_history: "ChatHistory", model: BaseArchytasModel, token_threshold: int) -> "list[RecordType]":
    from .chat_history import AIMessage
    token_count = chat_history.base_tokens or 0
    target_idx: int = -1
    summarizable_records = await get_summarizable_records(chat_history)
    for idx, record in enumerate(summarizable_records):
        token_count += record.token_count
        if token_count >= token_threshold:
            break
        target_idx = idx
    else:
        target_idx = -1

    # Don't split AIMessages and their tool call
    done = False
    while target_idx >= 0 and not done:
        target_record = summarizable_records[target_idx]
        target_message = target_record.message
        if isinstance(target_message, AIMessage) and target_message.tool_calls:
            target_idx -= 1
        else:
            done = True

    if target_idx >= 0:
        return summarizable_records[:target_idx+1]
    else:
        return []


async def get_records_up_to_loop(chat_history: "ChatHistory", model: BaseArchytasModel) -> "list[RecordType]":
    all_records = await get_summarizable_records(chat_history)
    loop_start_index = next((idx for idx, record in enumerate(all_records) if record.react_loop_id == chat_history.current_loop_id), None)
    if loop_start_index is not None:
        summarizable_records = all_records[:loop_start_index]
    else:
        summarizable_records = []
    return summarizable_records


async def default_loop_summarizer(
    loop_records: "list[MessageRecord]",
    chat_history: "ChatHistory",
    agent: "Agent",
    model: "BaseArchytasModel" = None,
    token_threshold: int = 4000,
    force_update: bool = False,
):
    from langchain_core.messages import ToolMessage
    coroutines = []
    for record in loop_records:
        if isinstance(record.message, ToolMessage) and (artifact := getattr(record.message, "artifact", None)):
            message = record.message
            tool_name = artifact.get("tool_name", None)
            tool_fn = agent.tools[tool_name]
            summarized: bool = artifact.get("summarized", False)
            summarizer: MessageSummarizerFunction = getattr(tool_fn, "summarizer", None)
            if summarizer and (force_update or not summarized):
                coroutines.append(summarizer(message, chat_history, agent, model=model))
    await asyncio.gather(*coroutines)


async def default_history_summarizer(
    chat_history: "ChatHistory",
    agent: "Agent",
    recordset: "list[MessageRecord[BaseMessage]|SummaryRecord]",
    force_update: bool = False,
):
    from .chat_history import MessageRecord, SummaryRecord, AIMessage, SystemMessage, HumanMessage, BaseMessage
    logger.debug(f"Summarizing history {chat_history=}, {agent=}, {force_update=}")

    if not recordset:
        return

    records_to_summarize: list[MessageRecord[BaseMessage]] = []
    summaries: list[SummaryRecord] = []
    for record in recordset:
        match record:
            case SummaryRecord():
                summaries.append(record)
            case MessageRecord():
                records_to_summarize.append(record)

    uuids = [record.uuid for record in records_to_summarize]

    jinja_globals = {
        "isinstance": isinstance,
        "MessageRecord": MessageRecord,
        "AIMessage": AIMessage,
        "SummaryRecord": SummaryRecord,
        "BaseMessage": BaseMessage,
    }
    system_template = jinja_env.get_template(
        name="summarization/default_history_summarization_system.jinja",
        globals=jinja_globals,
    )
    user_template = jinja_env.get_template(
        name="summarization/default_history_summarization_user.jinja",
        globals=jinja_globals
    )
    system_prompt = system_template.render({"chat_history": chat_history})
    user_prompt = user_template.render({
        "records_to_summarize": records_to_summarize,
        "summaries": summaries,
    })

    messages = [SystemMessage(content=system_prompt), HumanMessage(content=user_prompt)]
    response = await agent.model._model.ainvoke(
        input=messages,
    )
    summary_text = response.content
    usage_metadata = getattr(response, "usage_metadata", None)
    summary_tokens = usage_metadata.get("output_tokens")

    summary_record = SummaryRecord(
        message=SystemMessage(
            content=f"""\
Below is a summary of {len(uuids)} messages with UUIDs of: {uuids}

```summary
{summary_text}
```
""",
        ),
        summarized_messages=set(uuids),
        token_count=summary_tokens,
    )
    chat_history.summaries.append(summary_record)


async def default_tool_summarizer(message: "ToolMessage", chat_history: "ChatHistory", agent: "Agent"):
    message_length = len(message.content)
    if message_length < MESSAGE_SUMMARIZATION_THRESHOLD:
        # Message is already short
        return

    _, tool_call = chat_history.get_tool_caller(message.tool_call_id)

    message.content = f"""\
Summary of run:
ID: {tool_call.get("id")}
Name: {tool_call.get("name")}
Arguments: {json.dumps(tool_call.get("args"))}
Status: {message.status}
The first {MESSAGE_SUMMARIZATION_SNIPPET_SIZE} characters of the tools output are:
```
{message.content[:MESSAGE_SUMMARIZATION_SNIPPET_SIZE]}
```
"""
    message.artifact["summarized"] = True



async def llm_message_summarizer(message: "ToolMessage", chat_history: "ChatHistory", agent: "Agent"):
    from .chat_history import MessageRecord, AIMessage
    calling_record: MessageRecord[AIMessage]
    calling_record, tool_call = chat_history.get_tool_caller(message.tool_call_id)

    prompt: str = """\
You are a ReAct agent who has been conversing with the user.
The full history is growing large, and tool outputs can be large, so you will be helping by summarizing the output of a tool call.
Please format your output in a way that is concise, yet provides enough context in case you need to refer back to the tool call
in the future.
If you ever need the full output of the tool, there will be a tool available that will allow you to temporarily retrieve the full
output for the period in which it is needed.
"""
    query_parts: list[str] = []
    query_parts.append(
        f"""\
Below is the output of calling this tool:
    ToolCallId: {tool_call.get('id')}
    Name: {tool_call.get("name")}
    Arguments: {json.dumps(tool_call.get("args"))}
    Status: {message.status}.

Please summarize the output for the purpose of future reference in the conversational history.
""")
    if calling_record.message.text():
        query_parts.append(
            f"""\
To assist in the summarization, the your thought process when calling the tool was:
```
{calling_record.message.text()}
```
""")
    query_parts.append(
        f"""\
Here is the output generated by the tool:
```
{message.text()}
```
""")
    query = "\n".join(query_parts)
    logger.warning(f"`````summarization prompt`````\n{prompt}\n````` end `````")


    message.additional_kwargs["orig_content"] = message.content

    summarized_text = await agent.oneshot(prompt=prompt, query=query)
    logger.warning(summarized_text)
    #TODO replace text

    message.artifact["summarized"] = True



import asyncio
import copy
import functools
import hashlib
import inspect
import json
import logging
import uuid
from dataclasses import MISSING
from abc import ABC, abstractmethod
from dataclasses import dataclass, field as dataclass_field
from typing import TYPE_CHECKING,Callable, Collection, Optional, TypeVar, Generic, cast, Any, TypeAlias, Coroutine
from typing_extensions import Self

from langchain_core.messages import BaseMessage, HumanMessage, SystemMessage, ToolMessage, FunctionMessage, AIMessage, ToolCall
from pydantic import Field

from .exceptions import AuthenticationError, ExecutionError, ModelError, ContextWindowExceededError
from .models.base import BaseArchytasModel
from .utils import ensure_async

from .exceptions import AuthenticationError
from .summarizers import (
    MessageSummarizerFunction, LoopSummarizerFunction, HistorySummarizerFunction, # BaseHistorySummarizer, # DefaultHistoryManager
    default_history_summarizer, default_loop_summarizer
)


if TYPE_CHECKING:
    from .agent import Agent

logger = logging.getLogger(__name__)


def tool_hash(tools: list[any]) -> str:
    import hashlib
    hash_value = hashlib.sha1()
    for tool in tools:
        hash_value.update(str(tool).encode())
    return hash_value.hexdigest()


class ContextMessage(SystemMessage):
    """Simple wrapper around a system message to facilitate message disambiguation."""

class AutoContextMessage(ContextMessage):
    """An automatically updating context message that remains towards the top of the message list."""

    default_content: str = Field(exclude=True)
    content_updater: Callable[[], str] = Field(exclude=True)

    def __init__(self, default_content: str, content_updater: Callable[[], str], model: Optional[BaseArchytasModel] = None, **kwargs):
        kwargs.update({"default_content": default_content, "content_updater": content_updater})
        super().__init__(content=default_content, **kwargs)
        self._token_count = None
        self._model = model

    @property
    def content_hash(self):
        return hashlib.sha1(self.text().encode()).hexdigest()

    @property
    def token_count(self):
        return self._token_count

    async def update_content(self):
        orig_hash = self.content_hash
        result = await ensure_async(self.content_updater())
        self.content = result
        if self.content_hash != orig_hash and self._model:
            print("Need to update count")
            self._token_count = await self._model.get_num_tokens_from_messages([self])


MessageType = TypeVar("MessageType", bound=BaseMessage)

@dataclass
class AgentResponse:
    text: str
    tool_calls: list[ToolCall]


@dataclass
class MessageRecord(Generic[MessageType]):
    message: MessageType
    uuid: str = dataclass_field(default_factory=lambda: uuid.uuid4().hex)
    token_count: Optional[int] = dataclass_field(default=None)
    metadata: dict[str, Any] = dataclass_field(default_factory=lambda: {})
    react_loop_id: Optional[int] = dataclass_field(default=None)


@dataclass
class SummaryRecord(MessageRecord[SystemMessage]):
    message: SystemMessage
    summarized_messages: set[str] = dataclass_field(default_factory=lambda: set())


RecordType: TypeAlias = MessageRecord | SummaryRecord


@dataclass
class OutputMessage:
    content: str | list[str | dict]


@dataclass
class OutboundModel:
    provider: str
    model_name: str
    context_window: Optional[int]


@dataclass
class OutboundChatHistory:
    records: list[RecordType]
    system_message: Optional[MessageRecord[SystemMessage]]
    tool_token_usage_estimate: Optional[int]
    model: OutboundModel
    token_estimate: Optional[int]
    message_token_count: Optional[int]
    summary_token_count: Optional[int]
    overhead_token_count: Optional[int]
    summarization_threshold: Optional[int]


class ChatHistory:
    base_tokens: int
    current_loop_id: int|None
    raw_records: list[MessageRecord]
    summaries: list[SummaryRecord]
    system_message: Optional[MessageRecord[SystemMessage]]
    model: Optional[BaseArchytasModel]
    loop_summarizer: Optional[callable]
    history_summarizer: Optional[callable]
    summarization_threshold: int
    tool_token_estimate: int
    _tool_hash: str
    _token_estimate: int|None
    history_summarization_task: Optional[asyncio.Task]

    _current_context_id: int
    auto_context_message: Optional[AutoContextMessage]
    auto_update_context: bool

    def __init__(
        self,
        messages: Optional[Collection[BaseMessage]] = None,
        model: Optional[BaseArchytasModel] = None,
        loop_summarizer: Optional[callable] = default_loop_summarizer,
        history_summarizer: Optional[callable] = default_history_summarizer,
    ):
        self.base_tokens = 0
        self.current_loop_id = None
        self.raw_records = []
        self.summaries = []
        self.history_summarization_task = None
        self.system_message = None
        self.model = model
        self.tool_token_estimate = 0
        self._tool_hash = ""
        self._token_estimate = None
        self.loop_summarizer = loop_summarizer
        self.history_summarizer = history_summarizer
        if messages:
            self.raw_records.extend((
                MessageRecord(message=message) for message in messages
            ))

        if self.model is not None:
            self.summarization_threshold = self.model.default_summarization_threshold
        else:
            self.summarization_threshold = -1

        # use to generate unique ids for context messages
        self._current_context_id = 0

        # Initialize the auto_context_message to empty
        self.auto_context_message = None
        self.auto_update_context = False

    def set_system_message(
        self,
        system_message: SystemMessage|str
    ):
        if isinstance(system_message, str):
            system_message = SystemMessage(system_message)
        self.system_message = MessageRecord(
            message=system_message,
        )

    @property
    def auto_context_message_token_estimate(self) -> int:
        if self.auto_context_message:
            return self.auto_context_message.token_count
        else:
            return 0

    async def update_auto_context(self):
        coroutines: list[Coroutine] = []
        if self.auto_update_context:
            if self.auto_context_message:
                coroutines.append(self.auto_context_message.update_content())
            for message in self.raw_records:
                if isinstance(message.message, AutoContextMessage):
                    coroutines.append(message.message.update_content())
            if coroutines:
                await asyncio.gather(*coroutines)

    @property
    def token_overhead(self):
        return sum((value for value in (
            self.base_tokens,
            self.tool_token_estimate,
            self.auto_context_message_token_estimate,
        ) if isinstance(value, int)))

    def get_tool_caller(self, tool_call_id: str) -> tuple[MessageRecord, ToolCall]:
        calling_record, tool_call = next(
            (
                (record, tc) for record
                in self.raw_records
                for tc in getattr(record.message, "tool_calls", [])
                if tc.get('id', MISSING) == tool_call_id
            ),
            (None, None),
        )
        return (cast(MessageRecord, calling_record), tool_call)

    async def needs_summarization(self, model: Optional[BaseArchytasModel]=None):
        if model is not None:
            threshold = model.summarization_threshold
        elif self.model is not None:
            threshold = self.model.summarization_threshold
            model = self.model
        else:
            return False
        token_estimate = await self.token_estimate(model=model)
        return token_estimate > threshold

    async def summarize_loop(
        self,
        loop_records: Optional[list[MessageRecord]] = None,
        loop_record_id: Optional[int] = None,
        agent: "Agent" = None,
        force_update: bool = False
    ):
        if not loop_records:
            if not loop_record_id:
                loop_record_id: int = self.current_loop_id
            if not loop_record_id:
                logger.warning("Unable to determine intended loop to summarize. Neither loop_records nor loop_record_id provided, and no current .")
                return []
            loop_records: list[MessageRecord] = [
                record for record
                in await self.all_records(auto_update_context=False)
                if record.react_loop_id == loop_record_id
            ]
        if loop_records and self.loop_summarizer:
            logger.debug("Initiating loop summarization")
            await self.loop_summarizer(loop_records, self, agent, model=agent.model, force_update=force_update)
            logger.debug("Loop summarization completed")

    async def summarize_history(
            self,
            agent: "Agent" = None,
            token_threshold: Optional[int] = None,
            force_update: bool = False,
            in_loop: bool = False,
        ):

        # Return running summarization task if one is running to prevent duplicate runs
        if self.history_summarization_task is not None:
            return self.history_summarization_task

        if self.history_summarizer is None:
            return

        if token_threshold is None:
            token_threshold = agent.model.summarization_threshold

        def task_callback(task: asyncio.Task):
            print(f"Summarization task {task} completed.")
            self.history_summarization_task = None

        threshold = token_threshold or self.summarization_threshold
        from .summarizers import get_records_up_to_threshold, get_records_up_to_loop
        if in_loop:
            recordset = await get_records_up_to_loop(
                chat_history=self,
                model=agent.model,
            )
        else:
            recordset = await get_records_up_to_threshold(
                chat_history=self,
                model=agent.model,
                token_threshold=threshold,
            )

        # If not recordset, nothing to summarize, nothing left to do.
        if not recordset:
            return

        task = asyncio.create_task(
            self.history_summarizer(
                chat_history=self,
                agent=agent,
                recordset=recordset,
            )
        )
        task.add_done_callback(task_callback)
        self.history_summarization_task = task
        return task

    async def token_estimate(
        self,
        model: BaseArchytasModel,
        messages: Optional[list[RecordType]|list[MessageType]]=None,
        force_update: bool = False,
        tools: Optional[dict]=None,
        use_cache: bool=True,
    ) -> Optional[int]:
        """
        Does not update/process any AutoContextMessages. All messages should be updated prior to calling this method.
        By default, will calculate the summarized chat history. If you need the full history, pass in an appropriate
        list of messages to the messages argument, e.g. `self.token_estimate(messages=messages.raw_messages)`.
        """

        if messages is None:
            messages = await self.records(auto_update_context=False)

        message_records: list[RecordType] = []
        base_messages: list[BaseMessage] = []
        tool_content: Optional[str] = None

        if self.tool_token_estimate is None or (self._tool_hash and self._tool_hash != tool_hash(tools)):
            if tools:
                lc_tools = model.convert_tools(tuple(tools.items()))
            elif model.lc_tools:
                lc_tools = model.lc_tools
            if lc_tools and (not self.tool_token_estimate or self._tool_hash != tool_hash(lc_tools)):
                self._tool_hash = tool_hash(lc_tools)
                tool_content = {
                    tool.name: {
                        "name": tool.name,
                        "description": tool.description,
                        "arg_schema": tool.args_schema.model_json_schema(),
                    }
                    for tool in lc_tools
                }
                self.tool_token_estimate = await model.get_num_tokens_from_messages([HumanMessage(content=json.dumps(tool_content))])

        for item in messages:
            if isinstance(item, MessageRecord):
                message_records.append(item)
            elif isinstance(item, BaseMessage):
                base_messages.append(item)
            else:
                raise ValueError(f"Unable to handle message of type {item.__class__.__name__}")

        if message_records and base_messages:
                raise ValueError("It is invalid to mix ChatHistory MessageRecords and LangChain BaseMessages.")

        if message_records:
            token_estimate_sum = self.token_overhead
            for record in message_records:
                if force_update or record.token_count is None:
                    content = record.message.content
                    message = HumanMessage(content=content)
                    try:
                        message_token_est = await model.get_num_tokens_from_messages([message])
                    except NotImplementedError:
                        message_token_est = 0
                    record.token_count = message_token_est
                token_estimate_sum += record.token_count
            self._token_estimate = token_estimate_sum
            return token_estimate_sum

        if base_messages:
            try:
                token_estimate = await model.get_num_tokens_from_messages([base_messages])
            except NotImplementedError:
                token_estimate = 0
            token_estimate_sum = self.base_tokens + self.tool_token_estimate + token_estimate
            self._token_estimate = token_estimate_sum
            return token_estimate_sum

        self._token_estimate = None
        return None

    async def records(self, auto_update_context: bool = True) -> list[RecordType]:
        """
        Messages
        """
        records: list[RecordType] = []
        summarized_messages: set[str] = set()

        if self.system_message:
            records.append(self.system_message)

        if auto_update_context:
            await self.update_auto_context()

        if self.auto_context_message:
            records.append(
                MessageRecord(message=self.auto_context_message))

        for summary_record in self.summaries:
            records.append(summary_record)
            summarized_messages.update(summary_record.summarized_messages)

        for message_record in self.raw_records:
            if message_record.uuid not in summarized_messages:
                records.append(message_record)
        return records

    async def messages(self, auto_update_context: bool = True) -> list[MessageType]:
        records = await self.records(auto_update_context=auto_update_context)
        return [cast(MessageType, record.message) for record in records]

    async def all_records(self, auto_update_context: bool = True) -> list[RecordType]:
        messages: list[RecordType] = []
        if self.system_message:
            messages.append(self.system_message)

        if auto_update_context:
            await self.update_auto_context()

        if self.auto_context_message:
            messages.append(
                MessageRecord(message=self.auto_context_message))
        messages.extend(self.raw_records)
        return messages

    async def all_messages(self, auto_update_context: bool = True) -> list[MessageType]:
        records = await self.all_records(auto_update_context=auto_update_context)
        return [cast(MessageType, record.message) for record in records]

    def add_message(
        self,
        message: MessageType,
        token_count: Optional[int] = None,
    ):
        """Appends a message to the message list."""
        record = MessageRecord(message=message, token_count=token_count, react_loop_id=self.current_loop_id)
        self.raw_records.append(record)
        return record



import copy
import json
import os
from abc import ABC, abstractmethod
from pydantic import BaseModel as PydanticModel, ConfigDict, create_model, Field
from pydantic.fields import FieldInfo
from typing import TYPE_CHECKING, Annotated, Any, Optional, ClassVar, Sequence
from functools import cache

from langchain.tools import StructuredTool

if TYPE_CHECKING:
    from langchain_core.messages import AIMessage, BaseMessage
    from langchain_core.language_models.chat_models import BaseChatModel
    from ..agent import AgentResponse


class EnvironmentAuth:
    env_settings: dict[str, str]

    def __init__(self, **env_settings: dict[str, str]) -> None:
        for key, value in env_settings.items():
            if not (isinstance(key, str) and isinstance(value, str)):
                raise ValueError("EnvironmentAuth variables names and values must be strings.")
        self.env_settings = env_settings

    def apply(self):
        for env_name, env_value in self.env_settings.items():
            os.environ.setdefault(env_name, env_value)


def set_env_auth(**env_settings: dict[str, str]) -> None:
    for key, value in env_settings.items():
        if not (isinstance(key, str) and isinstance(value, str)):
            raise ValueError("EnvironmentAuth variables names and values must be strings.")
    for env_name, env_value in env_settings.items():
        os.environ.setdefault(env_name, env_value)


class ModelConfig(PydanticModel, extra='allow'):
    model_name: str
    model_config = ConfigDict(extra="allow", protected_namespaces=())
    api_key: str | None = None
    summarization_ratio: float | None = None
    summarization_threshold: int | None = None
    summarization_threshold_pct: int | None = None

    # extra fields --
    # max_tokens: int | None = None
    # region: str | None = None

class FinalAnswerSchema(PydanticModel):
    response: str = Field(..., description=(
        "Final response that should be displayed to the user, answering the user's question, summarizing the "
        "results of the task, and/or display any useful information. If any important information is returned by using "
        "a tool, be sure to include that information here as the user does not have access to the raw output of the "
        "tool executions."
    ))


class FailedTaskSchema(PydanticModel):
    reason: str = Field(...,
        description="A plain text explanation of the reason for the failure."
    )
    error: Optional[str] = Field(...,
        description=(
            "A plain text rendering of the underlying error, along with the stacktrace, if available, for debugging "
            "purposes."
        )
    )

final_answer = StructuredTool(
    name="final_answer",
    description="""\
This tool should ALWAYS be called last during a successful ReAct loop to provide a final response to the user. This
ensures that the user is properly informed as the user does not have access to all outputs from tools.
The response should either answer a question, summarize the results of a task, and/or provide any useful information
that the user will find helpful or desirable.
""",
    args_schema=FinalAnswerSchema,
)
fail_task = StructuredTool(
    name="fail_task",
    description=(
        "The fail_task tool is used to indicate that you have failed to complete the task. You should use this "
        "tool to communicate the reason for the failure to the user. Do not call this tool unless you have given "
        "a good effort to complete the task.\n"
        "In particular, you should call this tool if the same request keeps repeating itself and/or you do not "
        "seem to be able to make progress."
    ),
    args_schema=FailedTaskSchema,
)


class BaseArchytasModel(ABC):
    DEFAULT_MODEL: ClassVar[Optional[str]] = None
    DEFAULT_SUMMARIZATION_RATIO: float = 0.5
    MODEL_PROMPT_INSTRUCTIONS: str = ""

    _model: "BaseChatModel"
    config: ModelConfig
    lc_tools: "list[StructuredTool] | None"

    def __init__(self, config: ModelConfig | dict, **kwargs) -> None:
        if isinstance(config, dict):
            self.config = ModelConfig(**config)
        else:
            self.config = config
        self.auth(**kwargs)
        self._model = self.initialize_model(**kwargs)
        self.lc_tools = None

    def auth(self, **kwargs) -> None:
        pass

    @property
    def additional_prompt_info(self) -> str | None:
        return None

    @property
    def default_summarization_threshold(self) -> int | None:
        context_max = self.contextsize(self.model_name)
        if context_max is None:
            return None
        return int(context_max * self.DEFAULT_SUMMARIZATION_RATIO)

    @property
    def summarization_threshold(self) -> int | None:
        context_size = self.contextsize(self.model_name)
        if summarization_threshold := getattr(self.config, 'summarization_threshold', None):
            if context_size is None:
                return summarization_threshold
            else:
                return min(int(summarization_threshold), context_size)
        elif summarization_ratio := getattr(self.config, 'summarization_ratio', None):
            pass
        elif summarization_threshold_pct := getattr(self.config, 'summarization_threshold_pct', None):
            summarization_ratio = float(summarization_threshold_pct) / 100
            self.config.summarization_ratio = summarization_ratio
        else:
            summarization_ratio = self.DEFAULT_SUMMARIZATION_RATIO
        if context_size is None:
            return None
        return int(context_size * summarization_ratio)

    @property
    def model(self) -> "BaseChatModel":
        if self.lc_tools is not None:
            return self._model.bind_tools(self.lc_tools)
        else:
            return self._model

    @property
    def model_name(self) -> str | None:

        lc_model_name = getattr(self._model, "model", None)
        if isinstance(lc_model_name, str):
            return lc_model_name

        config_model_name = getattr(self.config, "model_name", None)
        if isinstance(config_model_name, str):
            return config_model_name

        class_default_modelname = getattr(self, "DEFAULT_MODEL", None)
        return class_default_modelname


    @cache
    def contextsize(self, model_name: Optional[str]=None) -> int | None:
        if model_name is None:
            model_name = self.model_name
        return None

    @abstractmethod
    def initialize_model(self, **kwargs):
        ...

    def invoke(self, input, *, config=None, stop=None, agent_tools=None, **kwargs):
        result = self.model.invoke(
            self._preprocess_messages(input),
            config,
            stop=stop,
            agent_tools=agent_tools,
            **kwargs
        )
        return result

    @staticmethod
    @cache
    def convert_tools(archytas_tools: tuple[tuple[str, Any], ...])-> "list[StructuredTool]":
        tools = [final_answer, fail_task]
        for name, tool in archytas_tools:
            arg_dict = {}
            for arg_name, arg_type, arg_desc, _ in tool._args_list:
                arg_dict[arg_name] = Annotated[arg_type.sub_type, FieldInfo(description=arg_desc)]
            if "thought" not in arg_dict:
                arg_dict["thought"] = Annotated[str, FieldInfo(description="Reasoning around why this tool is being called.")]
            tool_model = create_model(name, **arg_dict)
            lc_tool = StructuredTool(
                name=name,
                description=tool.__doc__,
                args_schema=tool_model,
                func=tool,
            )
            tools.append(lc_tool)
        return tools

    def set_tools(self, agent_tools: dict):
        agent_tools = tuple(
            sorted(
                [(name, func) for name, func in agent_tools.items() if not getattr(func, '_disabled', False)],
                key=lambda tool: tool[0]
            )
        )
        tools = self.convert_tools(agent_tools)
        self.lc_tools = tools
        return tools

    async def get_num_tokens_from_messages(
        self,
        messages: "list[BaseMessage]",
        tools: Optional[Sequence] = None,
    ) -> int:
        try:
            return self._model.get_num_tokens_from_messages(messages=messages, tools=tools)
        except Exception as err:
            print(err)
            pass
        return 0

    async def token_estimate(
        self,
        messages: "Optional[list[BaseMessage]]" = None,
        tools: Optional[dict] = None
    ):
        if tools:
            tools = tuple(
                sorted(
                    [(name, func) for name, func in tools.items() if not getattr(func, '_disabled', False)],
                    key=lambda tool: tool[0]
                )
            )
            tools = self.convert_tools(tools)
        messages = self._preprocess_messages(messages)
        messages: list[BaseMessage] = copy.deepcopy(messages)
        for message in messages:
            if isinstance(message.content, list):
                content = []
                for content_object in message.content:
                    match content_object:
                        case {"type": "text"}:
                            content.append(content_object)
                        case dict():
                            content.append(
                                {
                                    "type": "text",
                                    "text": json.dumps(content_object)
                                }
                            )
                        case _:
                            content.append(content_object)
                message.content = content

        return await self.get_num_tokens_from_messages(messages=messages, tools=tools)

    async def ainvoke(self, input, *, config=None, stop=None, agent_tools: dict=None, **kwargs):
        if self.lc_tools is None and agent_tools is not None:
            self.set_tools(agent_tools)

        try:
            messages = self._preprocess_messages(input)
            result = await self.model.ainvoke(
                messages,
                config,
                stop=stop,
                **kwargs
            )
            return result
        except Exception as error:
            print(error)
            return self.handle_invoke_error(error)

    def _preprocess_messages(self, messages: "list[BaseMessage]"):
        return messages

    def _rectify_result(self, response_message: "AIMessage"):
        return response_message

    def process_result(self, response_message: "AIMessage") -> "AgentResponse":
        from ..agent import AgentResponse
        content = response_message.content
        tool_calls = response_message.tool_calls
        tool_thoughts = [tool_call["args"].pop("thought", f"Calling tool '{tool_call['name']}'") for tool_call in tool_calls]

        match content:
            case list():
                text = "\n".join(item['text'] for item in content if item.get('type', None) == "text")
            case "":
                if tool_calls:
                    text = "\n".join(tool_thoughts)
                else:
                    raise ValueError("Response from LLM does not include any content or tool calls. This shouldn't happen.")
            case str():
                text = content
            case _:
                # TODO: Finish this
                raise ValueError("Response from LLM does not match expected format. Expected ")
        if text == "":
            text = "Thinking..."
        return AgentResponse(text=text, tool_calls=tool_calls)

    def handle_invoke_error(self, error: BaseException):
        raise error



import json
import re
import groq.resources
import requests
from functools import lru_cache
from typing import Optional, cast, ClassVar

import groq
from langchain_core.messages import BaseMessage, SystemMessage, HumanMessage, AIMessage, ToolMessage
from langchain_core.output_parsers import JsonOutputParser
from langchain_groq import ChatGroq
from pydantic import BaseModel as PydanticModel, Field

from .base import BaseArchytasModel, set_env_auth, ModelConfig

class Structure(PydanticModel):
    thought: str = Field(..., description="Thought process of why you are calling the tool")
    tool: str
    tool_input: str | dict | list
    helpful_thoguht: bool

class GroqModel(BaseArchytasModel):
    model: ChatGroq
    _client: groq.Groq
    api_key: str = ""

    DEFAULT_MODEL: ClassVar[str] = "llama3-8b-8192"
    MODEL_PROMPT_INSTRUCTIONS: str = """\
When generating JSON, remember to not wrap strings in triple quotes such as \'\'\' or \"\"\". If you want to add newlines \
to the JSON text, use `\\n` to add newlines.
Ensure all generated JSON is valid and would pass a JSON validator.
"""

    def __init__(self, config: ModelConfig) -> None:
        super().__init__(config)
        self._client = cast(groq.resources.chat.Completions, self.model.client)._client

    def auth(self, **kwargs) -> None:
        if 'api_key' in kwargs:
            self.api_key = kwargs['api_key']
        else:
            self.api_key = self.config.api_key
        if not self.api_key:
            raise ValueError("No auth credentials found.")

    def initialize_model(self, **kwargs):
        model = ChatGroq(
            model=self.config.model_name or self.DEFAULT_MODEL,
            api_key=self.api_key,
            base_url="https://api.groq.com/"
        )
        return model

    @property
    def model_name(self) -> str | None:
        model_name = getattr(self.model, "model_name", None)
        if model_name is not None:
            return model_name
        else:
            return getattr(self.config, "model_name", self.DEFAULT_MODEL)

    def ainvoke(self, input, *, config=None, stop=None, **kwargs):
        return super().ainvoke(input, config=config, stop=stop, **kwargs)

    def _preprocess_messages(self, messages: list[BaseMessage]):
        from ..agent import AutoContextMessage, ContextMessage
        output = []
        system_messages = []
        for message in messages:
            match message:
                case SystemMessage() | ContextMessage() | AutoContextMessage():
                    system_messages.append(message.content)
                case AIMessage():
                    # Duplicate mesage so we don't change raw storage
                    msg = message.model_copy()
                    if msg.tool_calls:
                        msg.tool_calls.clear()
                    output.append(msg)
                case _:
                    output.append(message)
        # Condense all context/system messages into a single first message as required by Anthropic
        output.insert(0, SystemMessage(content="\n".join(system_messages)))
        return output

    @lru_cache()
    def contextsize(self, model_name: Optional[str]=None) -> int | None:
        if model_name is None:
            model_name = self.model_name
        model_list = self._client.models.list()
        model_index = {model.id: model for model in model_list.data}
        model_info = model_index.get(model_name, None)
        context_window = getattr(model_info, "context_window", None)
        return context_window




class AuthenticationError(Exception):
    pass


class ModelError(Exception):
    pass


class ExecutionError(Exception):
    pass


class ContextWindowExceededError(ExecutionError):
    def __init__(self, *args, sent=None, maximum=None):
        super().__init__(*args)
        if isinstance(sent, str):
            self.sent = int(sent)
        if isinstance(maximum, str):
            self.maximum = int(maximum)



from archytas.react import ReActAgent, FailedTaskError
from archytas.tools import datetime_tool, timestamp, PythonTool
from archytas.tool_utils import tool
from archytas.demo_tools import (
    fib_n,
    example_tool,
    calculator,
    Jackpot,
    ModelSimulation,
    pirate_subquery,
)

from rich import print

from easyrepl import REPL


class CustomAgent(ReActAgent):
    """
    This ReAct agent has it's own tool(s) defined.
    """

    @tool()
    def where_am_i(self, prompt: str) -> str:
        """
        This tool is used is used if the user wants to know where they are located.

        Args:
            prompt (str): The prompt the user provided that caused this tool to be selected.

        Returns:
            str: The answer to the question asked.
        """
        return f"Why do you ask me '{prompt}'? You are on Earth."


def start_repl():
    # make an instance of a class tool
    jackpot = Jackpot(chips=1000)

    # make a list of the tools to use
    tools = [
        datetime_tool,
        timestamp,
        fib_n,
        example_tool,
        calculator,
        jackpot,
        ModelSimulation,
        PythonTool,
        pirate_subquery,
    ]

    # # example of making a python tool with a prelude and some pre-initialized local variables
    # import numpy as np
    # python = PythonTool(
    #     prelude='import numpy as np\nfrom matplotlib import pyplot as plt',
    #     locals={'fib_n': fib_n, 'jackpot': jackpot, 'ModelSimulation': ModelSimulation},
    # )
    # tools = [python]

    # create the agent
    agent = CustomAgent(tools=tools, verbose=True)

    # print the agent's prompt
    # print(agent.prompt)

    # run the REPL
    for query in REPL(history_file="chat_history.txt"):
        try:
            answer = agent.react(query)
            print(f"[green]{answer}[/green]")
        except FailedTaskError as e:
            print(f"[red]{e}[/red]")
        except KeyboardInterrupt:
            print("[yellow]KeyboardInterrupt[/yellow]")


if __name__ == "__main__":
    start_repl()



import asyncio
import logging
import os
from tenacity import (
    before_sleep_log,
    retry as tenacity_retry,
    retry_if_exception,
    stop_after_attempt,
    wait_exponential,
)
from typing import Callable, ContextManager, Any, Optional
from uuid import UUID

from rich import print as rprint
from rich.spinner import Spinner
from rich.live import Live


from langchain_core.messages import BaseMessage, HumanMessage, SystemMessage, ToolMessage, FunctionMessage, AIMessage, ToolCall

from .chat_history import ChatHistory, ContextMessage, AutoContextMessage, AgentResponse
from .exceptions import AuthenticationError, ExecutionError, ModelError
from .models.base import BaseArchytasModel

from .exceptions import AuthenticationError

logger = logging.getLogger(__name__)


def retry(fn):
    def is_openai_error(err):
        import openai
        # Don't retry auth errors. Assume to be permanent without fixing auth.
        if isinstance(err, openai.AuthenticationError):
            return False
        return isinstance(
            err,
            (
                openai.APITimeoutError,
                openai.APIError,
                openai.APIConnectionError,
                openai.RateLimitError,
                openai.InternalServerError,
            )
        )

    return tenacity_retry(
        reraise=True,
        stop=stop_after_attempt(4),
        wait=wait_exponential(multiplier=1, min=4, max=10),
        retry=retry_if_exception(is_openai_error),
        before_sleep=before_sleep_log(logger, logging.WARNING),
    )(fn)


logger = logging.getLogger(__name__)


def cli_spinner():
    return Live(
        Spinner("dots", speed=2, text="thinking..."),
        refresh_per_second=30,
        transient=True,
    )


class no_spinner:
    def __enter__(self):
        pass

    def __exit__(self, *args):
        pass


class Agent:
    chat_history: ChatHistory

    def __init__(
        self,
        *,
        model: BaseArchytasModel,
        prompt: str = "You are a helpful assistant.",
        api_key: str | None = None,
        spinner: Callable[[], ContextManager] | None = cli_spinner,
        rich_print: bool = True,
        verbose: bool = False,
        messages: Optional[list[BaseMessage]] | None = None,
        temperature: float = 0.0,
    ):
        """
        Agent class for managing communication with Language Models mediated by Langchain

        Args:
            model (BaseArchytasModel): The model to use. Defaults to OpenAIModel(model_name="gpt-4o").
            prompt (str, optional): The prompt to use when starting a new conversation. Defaults to "You are a helpful assistant.".
            api_key (str, optional): The LLM provider API key to use. Defaults to None. If None, the provider will use the default environment variable (e.g. OPENAI_API_KEY).
            spinner ((fn -> ContextManager) | None, optional): A function that returns a context manager that is run every time the LLM is generating a response. Defaults to cli_spinner which is used to display a spinner in the terminal.
            rich_print (bool, optional): Whether to use rich to print messages. Defaults to True. Can also be set via the DISABLE_RICH_PRINT environment variable.
            verbose (bool, optional): Expands the debug output. Includes full query context on requests to the LLM. Defaults to False.
            messages (list[BaseMessage], optional): A list of messages to initialize the agent's conversation history with. Defaults to an empty list.

        Raises:
            Exception: If no API key is given.
        """
        self.rich_print = bool(
            rich_print and not os.environ.get("DISABLE_RICH_PRINT", False)
        )
        self.verbose = verbose
        if not isinstance(model, BaseArchytasModel):
            # Importing OpenAI is slow, so limit import to only when it is needed.
            from .models.openai import OpenAIModel
            model_name = model if isinstance(model, str) else None
            self.model = OpenAIModel({"api_key": api_key, "model_name": model_name})
        else:
            self.model = model
        if not prompt:
            prompt = ""
        if hasattr(self.model, 'MODEL_PROMPT_INSTRUCTIONS'):
            prompt += "\n\n" + self.model.MODEL_PROMPT_INSTRUCTIONS
        self.chat_history = ChatHistory(messages)
        self.chat_history.set_system_message(SystemMessage(content=prompt))
        if spinner is not None and self.rich_print:
            self.spinner = spinner
        else:
            self.spinner = no_spinner

        self.temperature = temperature
        self.post_execute_task = None

    def print(self, *args, **kwargs):
        if self.rich_print:
            rprint(*args, **kwargs)
        else:
            print(*args, **kwargs)

    def debug(self, event_type: str, content: Any) -> None:
        """
        Debug handler

        Function at this level so it can be overridden in a subclass.
        """
        logger.debug(event_type, content)

    def log(self, event_type: str, content: Any) -> None:
        logger.info(event_type, content)

    def set_openai_key(self, key):
        import openai
        from .models.openai import OpenAIModel

        openai.api_key = key
        if isinstance(self.model, OpenAIModel):
            self.model.auth(api_key=key)


    async def all_messages(self) -> list[BaseMessage]:
        return await self.chat_history.all_messages()

    def add_context(self, context: str, *, lifetime: int | None = None) -> int:
        """
        Inject a context message to the agent's conversation.

        Useful for providing the agent with information relevant to the current conversation, e.g. tool state, environment info, etc.
        If a lifetime is specified, the context message will automatically be deleted from the chat history after that many steps.
        A context message can be deleted manually by calling clear_context() with the id of the context message.

        Args:
            context (str): The context to add to the agent's conversation.
            lifetime (Optional[int]): DEPRECATED. Ignored.

        Returns:
            int: The id of the context message.
        """
        if lifetime is not None:
            raise DeprecationWarning("Context message lifetimes have deprecated and will be ignored.")
        context_message = ContextMessage(
            content=context,
            lifetime=None,
        )
        record = self.chat_history.add_message(context_message)
        uuid = UUID(hex=record.uuid, version=4)
        return uuid.int

    def clear_context(self, id: int) -> None:
        """
        Remove a single context message from the agent's conversation.

        Args:
            id (int): The id of the context message to remove.
        """
        from .chat_history import MessageRecord
        idx: int = -1
        for index, record in enumerate(self.chat_history.raw_records):
            if isinstance(record, MessageRecord) and isinstance(record.message, ContextMessage):
                if UUID(hex=record.uuid, version=4).int == id:
                    idx = index
                    break
        if idx > -1:
            del self.chat_history.raw_records[idx]

    def clear_all_context(self) -> None:
        """Remove all context messages from the agent's conversation."""
        from .chat_history import MessageRecord
        to_remove = []
        for index, record in enumerate(self.chat_history.raw_records):
            if isinstance(record, MessageRecord) and isinstance(record.message, ContextMessage):
                to_remove.append(index)
        # Remove from the back to front so that removals don't change the indexs of subsequent items
        for idx in reversed(sorted(to_remove)):
            del self.chat_history.raw_records[idx]

    def set_auto_context(
        self,
        default_content: str,
        content_updater: Callable[[], str] | None = None,
        auto_update: bool = True,
    ):
        """
        A special type of context message that is always towards the top (but after the prompt and any system messages).
        This allows an agent to automatically update its context based on live conditions without having to call a tool every time.

        Args:
            default_content (str): The default message/content of the context if the content updater has not or cannot be run.
            content_updater (callable): A function/lambda that takes no arguments and returns a string. The returned string
                                        becomes the new context value.
            auto_update (boolean): If true, the context will be updated on every call. Otherwise, the context can be updated by
                                   calling `chat_history.auto_context_message.update_content()` when desired.
        """
        self.chat_history.auto_update_context = auto_update
        self.chat_history.auto_context_message = AutoContextMessage(
            default_content=default_content,
            content_updater=content_updater,
            model=self.model,
        )

    async def handle_message(self, message: BaseMessage):
        """Appends a message to the message list and executes."""
        self.chat_history.add_message(message)
        return await self.execute()

    async def query(self, message: str) -> str:
        """Send a user query to the agent. Returns the agent's response"""
        return await self.handle_message(HumanMessage(content=message))

    async def observe(self, observation: str, tool_name: str) -> str:
        """Send a system/tool observation to the agent. Returns the agent's response"""
        return await self.handle_message(FunctionMessage(type="function", content=observation, name=tool_name))

    async def inspect(self, query: str) -> str:
        """Send one-off system query that is not recorded in history"""
        return await self.execute([HumanMessage(content=query)])

    async def error(self, error: BaseMessage | str, drop_error: bool = True) -> str:
        """
        Send an error message to the agent. Returns the agent's response.

        Args:
            error (str): The error message to send to the agent.
            drop_error (bool, optional): If True, the error message and LLMs bad input will be dropped from the chat history. Defaults to `True`.
        """
        if not isinstance(error, BaseMessage):
            error = AIMessage(content=error)
        result = await self.handle_message(error)
        return result

    async def post_execute(self):
        if await self.chat_history.needs_summarization(model=self.model):
            await self.chat_history.summarize_history(agent=self)

    async def execute(
        self,
        additional_messages: list[BaseMessage] = None,
        tools=None,
        auto_append_response: bool = True
    ) -> AgentResponse:
        if additional_messages is None:
            additional_messages = []
        with self.spinner():
            records = await self.chat_history.records(auto_update_context=True)
            messages = [record.message for record in records] + additional_messages
            # TODO: Keep this here?
            token_estimate = await self.chat_history.token_estimate(model=self.model, tools=tools)
            print("Token estimate for query: ", token_estimate)
            if self.verbose:
                self.debug(event_type="llm_request", content=messages)
            raw_result = await self.model.ainvoke(
                input=messages,
                temperature=self.temperature,
                agent_tools=tools,
            )
            usage_metadata = getattr(raw_result, "usage_metadata", None)
            print("Actual usage for query: ", usage_metadata)

        response_token_count = await self.model.token_estimate(messages=[HumanMessage(content=raw_result.content)])

        # Add the raw result to history
        if auto_append_response:
            self.chat_history.add_message(self.model._rectify_result(raw_result), token_count=response_token_count)

        if isinstance(usage_metadata, dict):
            total_token_count = usage_metadata.get("total_tokens", None)
            if total_token_count > token_estimate:
                self.chat_history.base_tokens = total_token_count - token_estimate
        else:
            total_token_count = token_estimate + response_token_count

        # Return processed result
        result = self.model.process_result(raw_result)

        if self.verbose:
            self.debug(event_type="llm_response", content=result)

        def task_callback(task):
            self.post_execute_task = None
        if self.post_execute_task is None:
            task = asyncio.create_task(self.post_execute())
        task.add_done_callback(task_callback)
        self.post_execute_task = task

        return result

    async def oneshot(self, prompt: str, query: str, tools=None) -> str:
        """
        Send a user query to the agent. Returns the agent's response.
        This method ignores any previous conversation history, as well as the existing prompt.
        The output is the raw LLM text withouth any postprocessing, so you'll need to handle parsing it yourself.

        Args:
            prompt (str): The prompt to use when starting a new conversation.
            query (str): The user query to send to the agent.

        Returns:
            str: The agent's response to the user query.
        """
        if tools is None:
            tools = []
        with self.spinner():
            if self.verbose:
                self.debug(event_type="llm_oneshot", content=prompt)
            completion = await self.model.model.ainvoke(
                input=[
                    SystemMessage(content=prompt),
                    HumanMessage(content=query),
                ],
                temperature=self.temperature,
                tools=tools,
            )

        # return the agent's response
        result = completion.content
        if self.verbose:
            self.debug(event_type="llm_response", content=result)
        return result

    def all_messages_sync(self) -> list[BaseMessage]:
        """Synchronous wrapper around the asynchronous all_messages method."""
        return asyncio.run(self.chat_history.all_messages())

    def query_sync(self, message: str) -> str:
        """Synchronous wrapper around the asynchronous query method."""
        return asyncio.run(self.query(message))

    def observe_sync(self, observation: str) -> str:
        """Synchronous wrapper around the asynchronous observe method."""
        return asyncio.run(self.observe(observation))

    def inspect_sync(self, message: str) -> str:
        """Synchronous wrapper around the asynchronous inspect method."""
        return asyncio.run(self.inspect(message))

    def error_sync(self, error: str, drop_error: bool = True) -> str:
        """Synchronous wrapper around the asynchronous error method."""
        return asyncio.run(self.error(error, drop_error))

    def execute_sync(self) -> str:
        """Synchronous wrapper around the asynchronous execute method."""
        return asyncio.run(self.execute())

    def oneshot_sync(self, prompt: str, query: str) -> str:
        """Synchronous wrapper around the asynchronous oneshot method."""
        return asyncio.run(self.oneshot(prompt, query))



import asyncio
import inspect
import json
import sys
import logging
import traceback
import typing
import uuid
from langchain_core.messages import AIMessage, HumanMessage, ToolMessage, ToolCall

from .agent import Agent, BaseMessage, SystemMessage, AgentResponse
from .chat_history import ChatHistory, SummaryRecord, MessageRecord
from .exceptions import ContextWindowExceededError
from .prompt import build_prompt, build_all_tool_names
from .tools import ask_user
from .tool_utils import make_tool_dict, sanitize_toolname, tool, AgentRef
from .models.base import BaseArchytasModel
from .utils import extract_json, ensure_async


logger = logging.Logger("archytas")


class Undefined:
    ...


class FailedTaskError(Exception):
    def __init__(
            self,
            message: str,
            last_error: typing.Optional[BaseException] = None,
            tool_call_id: typing.Optional[str] = None,
            extra: typing.Optional[str] = None,
            *extra_args: object
        ) -> None:
        self.message = message
        self.last_error = last_error
        self.tool_call_id = tool_call_id
        self.extra = extra
        super().__init__(*extra_args)

    def __str__(self) -> str:
        err_str = f"""
Error message:
  {self.message}
"""
        if self.last_error:
            err_str += f"""
During processing, the following error was raised:
  {self.last_error}
"""
        if self.extra:
            err_str += f"""
Extra details:
  {self.extra}
"""
        return err_str


def catch_failure(fn):
    def handle_error(messages: list[BaseMessage], error: Exception, agent: "ReActAgent|None" = None):
        last_ai_message = None
        seen_tool_message_ids = set()
        for message in messages[::-1]:
            match message:
                case AIMessage():
                    last_ai_message = message
                    break
                case ToolMessage():
                    seen_tool_message_ids.add(message.id)
        if not last_ai_message:
            return
        if last_ai_message.tool_calls:
            missing_tool_ids = set(tool_call["id"] for tool_call in last_ai_message.tool_calls if tool_call["id"] not in seen_tool_message_ids)
            for missing_tool_id in missing_tool_ids:
                error_message = ToolMessage(
                    content=str(error),
                    tool_call_id=missing_tool_id,
                )
                messages.append(error_message)
                if agent:
                    agent.chat_history.add_message(error_message)

    async def inner_async(self: "ReActAgent", *args, **kwargs):
        try:
            return await ensure_async(fn(self, *args, **kwargs))
        except FailedTaskError as failed_task:
            handle_error(self.chat_history.raw_records, failed_task, self)
            raise
        except Exception as error:
        # TODO: Is it good to handle other errors here too? Probably, but flow can be tricky.
        # TODO: Would need to find dangling tools and generate tool messages for any dangling tool calls
            handle_error(self.chat_history.raw_records, error, self)
            raise

    return inner_async


class LoopController:
    PROCEED = 0
    STOP_SUCCESS = 1
    STOP_FATAL = 2

    state: int

    def __init__(self) -> None:
        self.reset()

    def set_state(self, new_value):
        self.state = new_value

    def reset(self):
        self.state = 0


class ReActAgent(Agent):
    def __init__(
        self,
        *,
        model: BaseArchytasModel | None = None,
        api_key: str | None = None,
        tools: list = None,
        allow_ask_user: bool = True,
        max_errors: int | None = 3,
        max_react_steps: int | None = None,
        thought_handler: typing.Callable | None = Undefined,
        messages: typing.Optional[list[BaseMessage]] | None = None,
        custom_prelude: str = None,
        **kwargs,
    ):
        """
        Create a ReAct agent

        Args:
            model (BaseArchytasModel): The model to use. Defaults to OpenAIModel(model_name="gpt-4o").
            api_key (str, optional): The LLM provider API key to use. Defaults to None. If None, the provider will use the default environment variable (e.g. OPENAI_API_KEY).
            tools (list): A list of tools to use. Defaults to None. If None, only the system tools (final_answer, fail_task) will be used.
            allow_ask_user (bool): Whether to include the ask_user tool, which allows the model to ask the user for clarification. Defaults to True.
            max_errors (int, optional): The maximum number of errors to allow during a task. Defaults to 3.
            max_react_steps (int, optional): The maximum number of steps to allow during a task. Defaults to infinity.
            thought_handler (function, optional): Hook to control logging/output of the thoughts made in the middle of a react loop. Set to None to disable, or leave default of Undefined to
                    print to terminal. Otherwise expects a callable function with the signature of `func(thought: str, tool_name: str, tool_input: str) -> None`.
            messages (list[BaseMessage], optional): A list of messages to start the agent with. Defaults to None.
            custom_prelude (str, optional): A custom prelude to use instead of the default.
                If provided, this will completely replace the default prelude.
        """
        # create a dictionary for looking up tools by name
        tools = tools or []
        if allow_ask_user:
            tools.append(ask_user)
        tools.append(self)
        self._raw_tools = tools
        self.tools = make_tool_dict(tools)
        self.current_query = None
        self.loop_messages: list[BaseMessage] = []
        self.custom_prelude = custom_prelude

        if thought_handler is Undefined:
            self.thought_handler = self.thought_callback
        else:
            self.thought_handler = thought_handler

        # check that the tools dict keys match the list of generated tool names
        names, keys = sorted(build_all_tool_names(tools)), sorted([*self.tools.keys()])
        assert (
            names == keys
        ), f"Internal Error: tools dict keys does not match list of generated tool names. {names} != {keys}"

        # create the prompt with the tools, and initialize the agent
        self.prompt = build_prompt(custom_prelude=custom_prelude)
        super().__init__(model=model, prompt=self.prompt, api_key=api_key, messages=messages, **kwargs)
        self.model.set_tools(self.tools)
        self.update_prompt()

        # react settings
        self.max_errors = max_errors or float("inf")
        self.max_react_steps = max_react_steps or float("inf")

        # number of errors and steps during current task
        self.errors = 0
        self.steps = 0

        # keep track of the last tool used (for error messages)
        self.last_tool_name = ""

    def update_prompt(self):
        self.prompt = build_prompt(custom_prelude=self.custom_prelude)
        model_prompt = self.model.MODEL_PROMPT_INSTRUCTIONS
        if model_prompt:
            self.prompt += "\n\n" + model_prompt
        self.chat_history.set_system_message(SystemMessage(content=self.prompt))

    def disable(self, *tool_names):
        if len(tool_names) == 0:
            return
        tool = None
        for tool_name in tool_names:
            if tool_name in self.tools:
                tool = self.tools[tool_name]
            elif "." in tool_name:
                matches = [name for name in self.tools.keys() if name.endswith(f".{tool_name}")]
                if len(matches) > 1:
                    raise ValueError(f"Ambiguous name: Multiple tools called '{tool_name}'")
                tool = self.tools[matches[0]]
            if tool is None:
                # If tool is not enabled/included, it can't be disabled, so skip.
                continue
            if inspect.ismethod(tool):
                setattr(tool.__func__, '_disabled', True)
            else:
                setattr(tool, '_disabled', True)

        if self.model:
            self.model.set_tools(self.tools)
        self.update_prompt()

    def thought_callback(self, thought: str, tool_name: str, tool_input: str) -> None:
        if self.verbose:
            # TODO: better coloring
            self.print(
                f"thought: {thought}\ntool: {tool_name}\ntool_input: {tool_input}\n"
            )

    def react(self, query: str, react_context:dict=None) -> str:
        """
        Synchronous wrapper around the asynchronous react_async method.
        """
        return asyncio.run(self.react_async(query, react_context))

    def handle_message(self, message):
        self.loop_messages.append(message)
        return super().handle_message(message)

    async def post_execute(self):
        # Prevent default behavior
        pass

    async def post_loop(self, skip_summarization: bool=False):
        if not skip_summarization:
            await self.chat_history.summarize_loop(agent=self)
            # Summarize full chat history after summarizing the loop, if needed.
            # This should always occur after loop summarization otherwise as token counts may be dramatically reduced
            # by the step above.
            if await self.chat_history.needs_summarization(model=self.model):
                await self.chat_history.summarize_history(agent=self)
        # Clear loop specific items
        self.chat_history.current_loop_id = None
        self.current_query = None

    @catch_failure
    async def react_async(self, query: str, react_context:dict=None) -> str:
        """
        Asynchronous react loop function.
        Continually calls tools until a satisfactory answer is reached.
        """

        # reset error and steps counter
        self.errors = 0
        self.steps = 0
        self.loop_messages = []

        action: AgentResponse | None = None
        reaction: AgentResponse | None = None

        # Set the current query for use in tools, auto context, etc
        self.current_query = query

        self.chat_history.current_loop_id = uuid.uuid4().int
        action = await self.handle_message(HumanMessage(content=query))

        controller = LoopController()

        # ReAct loop
        while True:
            # Start processing last loops reaction as a new action
            if reaction is not None:
                action = reaction

            if not action.tool_calls:
                if action.text:
                    try:
                        # Check to ensure the content isn't an tool call in the wrong place.
                        action_json = json.loads(action.text)
                        if isinstance(action_json, dict) and "id" in action_json and "name" in action_json and "args" in action_json:
                            action.tool_calls.append(action_json)
                    except json.JSONDecodeError:
                        # Presume content to be a final answer
                        action.tool_calls.append({
                            "id": uuid.uuid4().hex,
                            "name": "final_answer",
                            "args": {
                                "response": action.text
                            }
                        })

            if action.tool_calls:
                for tool_call in action.tool_calls:
                    tool_id = tool_call["id"]
                    tool_name = tool_call["name"]
                    tool_args = tool_call["args"]

                    # exit ReAct loop if agent says final_answer or fail_task
                    if tool_name == "final_answer":
                        self.debug(
                            event_type="react_final_answer",
                            content={
                                "final_answer": tool_args,
                            }
                        )
                        self.current_query = None
                        # Store final answer as response in message,
                        response = tool_args.get("response", None)
                        if not response:
                            # TODO: Handle this case
                            raise ValueError("The LLM provided an empty message for a final_answer. This is not valid.")
                        final_message = ToolMessage(
                            content=str(response),
                            tool_call_id=tool_id,
                        )
                        self.chat_history.add_message(final_message)
                        await self.post_loop()
                        return response
                    if tool_name == "fail_task":
                        reason = tool_args.get("reason", "No reason identified.")
                        error = tool_args.get("error", None)
                        await self.post_loop(skip_summarization=True)
                        raise FailedTaskError(
                            message=reason,
                            extra=error,
                            tool_call_id=tool_id
                        )

                    if self.thought_handler:
                        self.thought_handler(action.text, tool_name, tool_args)


                    # run tool
                    try:
                        tool_fn = self.tools[tool_name]
                    except KeyError:
                        self.chat_history.add_message(ToolMessage(
                            content=f'Unknown tool "{tool_name}"\nAvailable tools: {", ".join(self.tools.keys())}',
                            tool_call_id=tool_id,
                        ))
                        continue

                    try:
                        tool_context = {
                            "agent": self,
                            "tool_name": tool_name,
                            "raw_tool": tool_fn,
                            "loop_controller": controller,
                            "react_context": react_context,
                        }
                        tool_self_ref = getattr(tool_fn, "__self__", None)
                        self.log(
                            event_type="react_tool",
                            content={
                                "tool": tool_name,
                                "input": tool_args,
                            }
                        )
                        tool_output = await tool_fn.run(args=tool_args, tool_context=tool_context, self_ref=tool_self_ref)
                        self.log(
                            event_type="react_tool_output",
                            content={
                                "tool": tool_name,
                                "input": tool_args,
                                "output": tool_output,
                            }
                        )

                        tool_message = ToolMessage(
                            content=tool_output,
                            tool_call_id=tool_id,
                            artifact={
                                "tool_name": tool_name,
                                "summarized": False
                            }
                        )

                        # Always add tool response before handling state changes to ensure message history is correct.
                        self.chat_history.add_message(tool_message)

                        # Have the agent observe the result, and get the next action
                        if self.verbose:
                            self.display_observation(tool_output)

                        # Log cases when tools override controller state
                        if controller.state != LoopController.PROCEED:
                            self.debug(
                                event_type="react_controller_state",
                                content={
                                    "state": controller.state,
                                }
                            )
                        # Check loop controller to see if we need to stop or error
                        if controller.state == LoopController.STOP_SUCCESS:
                            await self.post_loop()
                            return tool_output
                        if controller.state == LoopController.STOP_FATAL:
                            await self.post_loop()
                            raise FailedTaskError(
                                tool_output,
                                tool_call_id=tool_id
                            )
                    except asyncio.CancelledError:
                        self.chat_history.add_message(ToolMessage(
                            content='Execution of this tool was interrupted by the user.',
                            tool_call_id=tool_id
                        ))
                        self.chat_history.current_loop_id = None
                    except Exception as e:
                        self.chat_history.add_message(ToolMessage(
                            content=f'error running tool "{tool_name}"\n\n:{e}\n{traceback.format_exception(e)}',
                            tool_call_id=tool_id
                        ))
                        self.chat_history.current_loop_id = None
                        if self.verbose:
                            self.print(f"[red]Exception raised in tool: '{tool_name}'[/red]\n[yellow]{''.join(traceback.format_exception(e))}[/yellow]")

            # Execute to fetch next step in the ReAct loop
            try:
                reaction = await self.execute()
            except ContextWindowExceededError:
                # Fetch reference
                history_summarization_task = self.chat_history.history_summarization_task
                if not history_summarization_task:
                    # This will kick off a task, tracked via the history_summarization_task variable on chat_history
                    history_summarization_task = await self.chat_history.summarize_history(agent=self, in_loop=True)
                # Wait for summarization task to complete.
                if history_summarization_task:
                    await history_summarization_task
                # Now that summarization has completed, try to execute again.
                reaction = await self.execute()

    @staticmethod
    def extract_action(action: dict) -> tuple[str, str, str, bool]:
        """Verify that action has the correct keys. Otherwise, raise an error"""
        assert isinstance(
            action, dict
        ), f"Action must be a json dictionary, got {type(action)}"
        assert "thought" in action, "Action json is missing key 'thought'"
        assert "tool" in action, "Action json is missing key 'tool'"
        assert "tool_input" in action, "Action json is missing key 'tool_input'"

        thought = action["thought"]
        tool = action["tool"]
        tool_input = action["tool_input"]
        helpful_thought = action.get("helpful_thought", True)


        return thought, tool, tool_input, helpful_thought

    def execute(self, additional_messages: list[BaseMessage] = None) -> str:
        """
        Execute the model and return the output (see `Agent.execute()`).
        Keeps track of the number of execute calls, and raises an error if there are too many.
        """
        if additional_messages is None:
            additional_messages = []
        self.steps += 1
        if self.steps > self.max_react_steps:
            last_tool_id = None
            raise FailedTaskError(
                f"Too many steps ({self.steps} > max_react_steps) during task.\nLast action should have been either final_answer or fail_task. Instead got: {self.last_tool_name}",
                tool_call_id=last_tool_id,
            )
        return super().execute(additional_messages, tools=self.tools)

    def error(self, mesg: str, err: BaseException, tool_id: str | None = None) -> str:
        """error handling. If too many errors, break the ReAct loop. Otherwise tell the agent, and continue"""
        self.errors += 1
        if self.errors >= self.max_errors:
            raise FailedTaskError(
                f"Too many errors during task. Last error: {mesg}",
                last_error=err,
                tool_call_id=tool_id
            )
        if self.verbose:
            if self.rich_print:
                self.print(f"[red]error: {mesg}[/red]", file=sys.stderr)
            else:
                self.print(f"error: {mesg}", file=sys.stderr)

    def display_observation(self, observation):
        """
        Display the observation. Can be overridden by subclasses to display the observation in different ways.
        """
        self.print(f"observation: {observation}\n")

    @tool(autosummarize=True)
    def retrieve_summarized_messages_of_summary(
        self,
        summary_record_uuid: str,
        agent_ref: AgentRef,
    ) -> str:
        """
        Temporarily retrieves and rehydrates the full, raw contents of the messages summarized by the referenced record.
        The contents will be available during the current ReAct loop, but will be removed once the current loop is
        finished.

        Args:
            summary_record_uuid (str): The UUID value of the summary record from which you want to recover the
                    full, unsummarized messages.

        Returns:
            str: The full contents of the requested previously summarized messages.
"""
        from typing import cast
        agent = cast(ReActAgent, agent_ref)
        summary: SummaryRecord

        summary = next(
            (
                summary_record for summary_record
                in agent.chat_history.summaries
                if summary_record.uuid == summary_record_uuid
                or summary_record_uuid in summary_record.summarized_messages
            ),
            None
        )

        if not summary:
            return f"Error: Unable to locate summary {summarized_record_uuid}"

        # This method of collecting records should ensure correct ordering
        summarized_records: list[MessageRecord[BaseMessage]] = [
            message_record for message_record
            in agent.chat_history.raw_records
            if message_record.uuid in summary.summarized_messages
        ]

        if len(summarized_records) != len(summary.summarized_messages):
            return f"Error: Unable to retrieve all messages from summary {summary.uuid}"

        if not summarized_records:
            all_summaries = [
                (summary_record.uuid, summary_record.summarized_messages)
                for summary_record in agent.chat_history.summaries
            ]
            if all_summaries:
                summary_text = "\n".join([f"  Summary `{summary_uuid}`:\n    Message uuids: {', '.join(messages)}" for summary_uuid, messages in all_summaries])
            else:
                summary_text = "  No summaries exist."
            response = f"""\
Error: Unable to locate a the requested records.
Available summaries are:
    {summary_text}
"""
            return response

        response_parts: list[str] = []
        response_parts.append(f"""Messages summarized by summary record with UUID {summary.uuid}:""")

        for record in summarized_records:
            response_parts.append(f"""\
==== Message `{record.uuid}`     ====
--- Message text     ---
{record.message.text()}
--- End Message text ---
"""
            )
            if isinstance(record.message, AIMessage) and record.message.tool_calls:
                for tool_call in record.message.tool_calls:
                    response_parts.extend([
                        f"""\
--- Tool Call     ---
""",
                        json.dumps(tool_call),
                        f"""\
--- End Tool Call ---
""",
                    ])
            response_parts.append(f"""\
==== END Message `{record.uuid}` ====
                """)
        return "\n".join(response_parts)



import os
import re
from functools import lru_cache
from typing import Any, Optional, Annotated
from langchain_openai.chat_models import ChatOpenAI
from langchain_openai.llms.base import OpenAI
from langchain_core.messages import FunctionMessage, AIMessage
from langchain.tools import StructuredTool

from openai import AuthenticationError as OpenAIAuthenticationError, APIError, APIConnectionError, RateLimitError, OpenAIError, BadRequestError
from .base import BaseArchytasModel, ModelConfig, set_env_auth
from ..exceptions import AuthenticationError, ExecutionError, ContextWindowExceededError

import logging
logger = logging.getLogger(__name__)

DEFERRED_TOKEN_VALUE = "***deferred***"

class OpenAIModel(BaseArchytasModel):
    DEFAULT_MODEL = "gpt-4o"

    def __init__(self, config, **kwargs):
        super().__init__(config, **kwargs)

    def auth(self, **kwargs) -> None:
        auth_token = None
        if 'api_key' in kwargs:
            auth_token = kwargs['api_key']
        else:
            auth_token = self.config.api_key
        if not auth_token:
            auth_token = DEFERRED_TOKEN_VALUE
        set_env_auth(OPENAI_API_KEY=auth_token)

        # Replace local auth token from value from environment variables to allow fetching preset auth variables in the
        # environment.
        auth_token = os.environ.get('OPENAI_API_KEY', DEFERRED_TOKEN_VALUE)

        if auth_token != DEFERRED_TOKEN_VALUE:
            self.config.api_key = auth_token
        # Reset the openai client with the new value, if needed.
        if getattr(self, "model", None):
            self.model.openai_api_key._secret_value = auth_token
            self.model.client = None
            self.model.async_client = None

            # This method reinitializes the clients
            self.model.validate_environment()

    def initialize_model(self, **kwargs):
        try:
            model = self.config.model_name or self.DEFAULT_MODEL
            titoken_model_name =  "gpt-4o" if 'gpt-4.1' in model else model
            return ChatOpenAI(model=model, tiktoken_model_name=titoken_model_name)
        except (APIConnectionError, OpenAIError) as err:
            if not self.config.api_key:
                raise AuthenticationError("OpenAI API Key not set")
            else:
                raise AuthenticationError("OpenAI Authentication Error") from err

    def ainvoke(self, input, *, config=None, stop=None, **kwargs):
        if self.model is None or not getattr(self.model, 'openai_api_key', None):
            raise AuthenticationError("OpenAI API Key missing")
        if "o3" in self.model.model_name.lower():
            # o3 doesn't accept a temperature keyword on invoke
            kwargs.pop("temperature")
        return super().ainvoke(input, config=config, stop=stop, **kwargs)

    def handle_invoke_error(self, error: BaseException):
        if isinstance(error, OpenAIAuthenticationError):
            raise AuthenticationError("OpenAI Authentication Error") from error
        elif isinstance(error, RateLimitError):
            raise ExecutionError(error.message) from error
        elif isinstance(error, (APIConnectionError, OpenAIError)) and not self.model.openai_api_key:
            raise AuthenticationError("OpenAI Authentication Error") from error
        elif isinstance(error, (BadRequestError)) and error.code == "context_length_exceeded":
            raise ContextWindowExceededError(error.body.get('message', None)) from error
        else:
            raise error

    @lru_cache()
    def contextsize(self, model_name = None):
        if model_name is None:
            model_name = self.model_name
        try:
            return OpenAI.modelname_to_contextsize(model_name)
        except ValueError as err:
            if 'gpt-4.1' in model_name:
                return 1_000_000
            elif model_name.startswith(('o3', 'o4')):
                return 200_000
            raise



import json
import logging
import os

from botocore.exceptions import ClientError, NoCredentialsError
from langchain_aws import ChatBedrockConverse
from langchain_core.messages import SystemMessage

from archytas.agent import BaseMessage

from ..exceptions import AuthenticationError
from .base import BaseArchytasModel, ModelConfig

DEFERRED_TOKEN_VALUE = "***deferred***"

class BedrockModel(BaseArchytasModel):
    # foundation model names are the last part of the ARN after the slash
    # but we have to use inference profiles
    DEFAULT_MODEL: str = "us.anthropic.claude-3-5-sonnet-20241022-v2:0"
    DEFAULT_REGION: str = "us-east-1"

    tool_name_map: dict
    rev_tool_name_map: dict

    credentials_profile_name: str | None
    aws_access_key: str | None
    aws_secret_key: str | None
    aws_session_token: str | None
    region: str

    def __init__(self, config: ModelConfig, **kwargs) -> None:
        self.last_messages: list[BaseMessage] | None = None
        self.tool_name_map = {}
        self.rev_tool_name_map = {}
        self.credentials_profile_name = None
        self.aws_access_key = None
        self.aws_secret_key = None
        self.aws_session_token = None
        super().__init__(config, **kwargs)

    def auth(self, **kwargs) -> None:
        if self.config.model_extra:
            access = self.config.model_extra.get('aws_access_key', '')
            secret = self.config.model_extra.get('aws_secret_key', '')
            session = self.config.model_extra.get('aws_session_token', '')
            if access != '' and secret != '':
                self.aws_access_key = access
                self.aws_secret_key = secret
                self.aws_session_token = session if session != '' else None


        # not handled - running on EC2 and expecting to authenticate via instance profile and IMDSv2
        # TODO: handle ec2/instance profile if we need it later. could be as easy as removing the exception below
        if 'credentials_profile_name' in kwargs:
            self.credentials_profile_name = kwargs['credentials_profile_name']
            return

        # required if not using credentials or env vars. manually passing the argument should take
        # precedence over env vars
        aws_keys = ['aws_access_key', 'aws_secret_key']
        if any([(key in kwargs) for key in aws_keys + ['aws_session_token']]):
            for key in aws_keys:
                if key in kwargs:
                    setattr(self, key, kwargs['key'])
                else:
                    raise AuthenticationError(f'one of aws_access_key or aws_secret_key was missing: Missing key: {key}')
            # NOT required, but if present, the above two also *must* exist.
            self.aws_session_token = kwargs.get('aws_session_token', None)
            return

        env_vars = {
            'AWS_ACCESS_KEY_ID': 'aws_access_key',
            'AWS_SECRET_ACCESS_KEY': 'aws_secret_key'
        }
        if any([(var in os.environ) for var in list(env_vars.keys()) + ['AWS_SESSION_TOKEN']]):
            for var in env_vars:
                if var in os.environ:
                    setattr(self, env_vars[var], os.environ[var])
                else:
                    raise AuthenticationError(f'missing one of required env vars: access_key or secret_key: {var}')
            self.aws_session_token = os.environ.get('AWS_SESSION_TOKEN', None)


    def initialize_model(self, **kwargs):
        region = os.environ.get('AWS_REGION', self.DEFAULT_REGION)
        max_tokens = None
        if self.config.model_extra:
            if self.config.model_extra.get('region', '') != '':
                region = self.config.model_extra.get('region')
            max_tokens = self.config.model_extra.get('max_tokens', 4096)

        model = self.config.model_name or self.DEFAULT_MODEL

        if self.credentials_profile_name:
            return ChatBedrockConverse(
                credentials_profile_name=self.credentials_profile_name,
                region_name=region,
                model=model,
                max_tokens=max_tokens
            )
        else:
            return ChatBedrockConverse(
                aws_access_key_id=self.aws_access_key,
                aws_secret_access_key=self.aws_secret_key,
                aws_session_token=self.aws_session_token or None,
                region_name=region,
                model=model,
                max_tokens=max_tokens or 4096
            )

    def _preprocess_messages(self, messages):
        from ..agent import AutoContextMessage, ContextMessage
        output = []

        system_messages = []
        # Combine all system/context/autocontext messages into a single initial system message
        for message in messages:
            match message:
                case SystemMessage() | ContextMessage() | AutoContextMessage():
                    system_messages.append(message.content)
                case _:
                    output.append(message)
        # Condense all context/system messages into a single first message as required by Anthropic
        output.insert(0, SystemMessage(content="\n".join(system_messages)))
        self.last_messages = [msg.model_copy(deep=True) for msg in output]
        return output

    def handle_invoke_error(self, error: BaseException):
        # client error catches a lot of credentials errors like incorrect profile names
        if isinstance(error, NoCredentialsError) or (isinstance(error, ClientError) and error.response["ResponseMetadata"]["HTTPStatusCode"] == 403):
            raise AuthenticationError(f"{error}")
        # TODO: Retry with delay on rate limit errors?
        # elif isinstance(error, RateLimitError):
        #     raise
        else:
            if self.last_messages:
                message_output = [msg.model_dump() for msg in self.last_messages]
                logging.warning(
                    "An exception has occurred. Below are the messages that were sent to in the most recent request:\n" +
                    json.dumps(message_output, indent=2)
                )
            raise



import inspect
import json
import re
from json.decoder import JSONDecoder
from types import GenericAlias, UnionType
from typing import Any, Union, get_origin, get_args as get_type_args, Coroutine, Callable

json_regex = re.compile(r'(```)(json)?(.*?)(```)')

def extract_json(text: str) -> dict | list:
    """
    Finds and extracts JSON from a block of text, probably a response from a LLM.
    """
    from langchain_core.utils.json import parse_json_markdown
    # Take advantage of the functionality in langchain
    result = parse_json_markdown(text)
    return result


def get_local_name(val: Any, locals: dict[str, Any]) -> str:
    """
    Determines the name of a local variable by finding a matching value in the given locals dict

    Performs a linear search over locals, so may not be performant if done frequently.

    Args:
        val (Any): The value to find the name of. This can be a module, class, function, variable, etc.
        locals (dict[str, Any]): The locals dict to search for the value in. This is usually the locals() dict from the root scope.

    Returns:
        str: The name of the local variable

    Raises:
        ValueError: If the value is not found in the locals dict
    """
    for name, obj in locals.items():
        if obj is val:
            return name
    raise ValueError(f"Value {val} not found in locals")


class InstanceMethod:
    """
    A descriptor that wraps a method so that it can be accessed from an instance of the class, but not from the class itself.
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, instance, owner):
        if instance is None:
            raise TypeError(
                "This method should only be accessed from an instance of the class"
            )
        return self.func.__get__(instance, owner)


async def ensure_async(fn: Coroutine|Callable):
    if inspect.iscoroutine(fn) or inspect.iscoroutinefunction(fn):
        return await fn
    else:
        return fn



from langchain_openai import AzureChatOpenAI
from archytas.exceptions import AuthenticationError
from archytas.models.base import ModelConfig, set_env_auth
from openai import AuthenticationError as OpenAIAuthenticationError, APIError, APIConnectionError, RateLimitError, OpenAIError
from .openai import OpenAIModel
import os

DEFERRED_TOKEN_VALUE = "***deferred***"

class AzureOpenAIModel(OpenAIModel):
    def __init__(self, config: ModelConfig, **kwargs):
        super().__init__(config, **kwargs)

    def auth(self, **kwargs) -> None:
        auth_token = None
        if 'api_key' in kwargs:
            auth_token = kwargs['api_key']
        else:
            auth_token = self.config.api_key
        if not auth_token:
            auth_token = DEFERRED_TOKEN_VALUE
        set_env_auth(AZURE_OPENAI_API_KEY=auth_token)

        # Replace local auth token from value from environment variables to allow fetching preset auth variables in the
        # environment.
        auth_token = os.environ.get('AZURE_OPENAI_API_KEY', DEFERRED_TOKEN_VALUE)

        if auth_token != DEFERRED_TOKEN_VALUE:
            self.config.api_key = auth_token
        # Reset the openai client with the new value, if needed.
        if getattr(self, "model", None):
            self.model.azure_openai_api_key._secret_value = auth_token
            self.model.client = None
            self.model.async_client = None

            # This method reinitializes the clients
            self.model.validate_environment()

    def initialize_model(self, **kwargs):
        if 'AZURE_OPENAI_ENDPOINT' in os.environ:
            self.endpoint = os.environ.get('AZURE_OPENAI_ENDPOINT')
        elif self.config.model_extra is None or 'endpoint' not in self.config.model_extra:
            raise AuthenticationError('Azure OpenAI models must have endpoint set.')
        else:
            self.endpoint = self.config.model_extra['endpoint']
        try:
            return AzureChatOpenAI(
                model=self.config.model_name or "",
                azure_endpoint=self.endpoint,
                api_version='2024-10-21'
            )
        except (APIConnectionError, OpenAIError) as err:
            if not self.config.api_key:
                raise AuthenticationError("OpenAI API Key not set")
            else:
                raise AuthenticationError("OpenAI Authentication Error") from err


class AzureFoundryModel():
    pass



import json
import re
import logging
from functools import lru_cache

from anthropic import AuthenticationError as AnthropicAuthenticError, RateLimitError, BadRequestError
from langchain_anthropic.chat_models import ChatAnthropic
from langchain_core.messages import AIMessage, SystemMessage
from langchain_anthropic.llms import AnthropicLLM
from pydantic import BaseModel as PydanticModel, Field

from archytas.agent import AIMessage, BaseMessage

from .base import BaseArchytasModel, EnvironmentAuth, ModelConfig
from ..exceptions import AuthenticationError, ExecutionError, ContextWindowExceededError

class DummyTool(PydanticModel):
    """
    Dummy Tool
    """
    input: str = Field(..., description="input")


class AnthropicModel(BaseArchytasModel):
    DEFAULT_MODEL: str = "claude-3-5-sonnet-latest"
    api_key: str = ""
    tool_name_map: dict
    rev_tool_name_map: dict

    def __init__(self, config: ModelConfig, **kwargs) -> None:
        super().__init__(config, **kwargs)
        self.tool_name_map = {}
        self.rev_tool_name_map = {}

    def auth(self, **kwargs) -> None:
        if 'api_key' in kwargs:
            self.api_key = kwargs['api_key']
        else:
            self.api_key = self.config.api_key or ""
        if not self.api_key:
            raise AuthenticationError("No auth credentials found.")

    def initialize_model(self, **kwargs):
        max_tokens = 4096
        if self.config.model_extra:
            max_tokens = self.config.model_extra.get('max_tokens', 4096)
        return ChatAnthropic(
            model=self.config.model_name or self.DEFAULT_MODEL,
            api_key=self.api_key,
            max_tokens=max_tokens
        )

    def _preprocess_messages(self, messages):
        from ..agent import AutoContextMessage, ContextMessage
        output = []

        system_messages = []
        # Combine all system/context/autocontext messages into a single initial system message
        for message in messages:
            match message:
                case SystemMessage() | ContextMessage() | AutoContextMessage():
                    system_messages.append(message.content)
                case _:
                    output.append(message)
        # Condense all context/system messages into a single first message as required by Anthropic
        output.insert(0, SystemMessage(content="\n".join(system_messages)))
        return output

    def handle_invoke_error(self, error: BaseException):
        if isinstance(error, AnthropicAuthenticError):
            raise AuthenticationError("Anthropic Authentication Error") from error
        # TODO: Retry with delay on rate limit errors?
        # elif isinstance(error, RateLimitError):
        #     raise
        elif isinstance(error, BadRequestError) and "prompt is too long" in error.message:
            sent = None
            maximum = None
            try:
                body = error.response.json().get("error", {})
                if "message" in body:
                    match: re.Match = re.search(r'prompt is too long: (\d+) .* (\d+) maximum', body["message"])
                    if match:
                        sent, maximum = match.groups()
            finally:
                raise ContextWindowExceededError(*error.args, sent=sent, maximum=maximum) from error
        else:
            # if self.last_messages:
            #     message_output = [msg.model_dump() for msg in self.last_messages]
            #     logging.warning(
            #         "An exception has occurred. Below are the messages that were sent to in the most recent request:\n" +
            #         json.dumps(message_output, indent=2)
            #     )
            raise

    @lru_cache()
    def contextsize(self, model_name = None):
        if model_name is None:
            model_name = self.model_name
        # TODO: This is accurate for all current models (as of 2025-02-27), for all Haiku and Sonnet models. Seems like
        # new models would have a new name if the context window changes. Hopefully in the future, this value can be
        # retrieved programatically.
        if model_name and ("haiku" in model_name or "sonnet" in model_name):
            return 200_000
        else:
            return None
